@@ -1,10 +1,9 @@
 /* update-passwd - Safely update /etc/passwd, /etc/shadow and /etc/group
- * Copyright (C) 1999,2000 Wichert Akkerman
+ * Copyright 1999-2002 Wichert Akkerman <wichert@deephackmode.org>
  *
  * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2, or (at your option)
- * any later version.
+ * it under the terms of version 2 of the GNU General Public License as
+ * published by the Free Software Foundation.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -15,9 +14,6 @@
  * along with this program; if not, write to the Free Software Foundation,
  * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
  *
- * Written by Wichert Akkerman <wakkerma@debian.org>, based on an earlier
- * attempt from Galen Hazelwood <galenh@micron.net>.
- *
  */
 
 #ifdef HAVE_CONFIG_H
@@ -40,7 +36,7 @@
 #include <shadow.h>
 #include <grp.h>
 
-#define VERSION			"3.1.8"
+#define VERSION			"3.4.1"
 
 #define DEFAULT_PASSWD_MASTER	"/usr/share/base-passwd/passwd.master"
 #define DEFAULT_GROUP_MASTER	"/usr/share/base-passwd/group.master"
@@ -48,7 +44,7 @@
 #define	WRITE_EXTENSION		".upwd-write"
 #define	BACKUP_EXTENSION	".org"
 
-#define LINESIZE		1024
+#define OUR_NSS_BUFSIZE		8192
 
 
 #define FL_KEEPHOME	0x0001
@@ -60,26 +56,42 @@
 #define FL_NOAUTOADD	0x0020
 
 /* This structure is actually used for both users and groups
- * we probably should split that sometime.
+ * we probably should split that someday.
  */
-const struct _userinfo {
-    unsigned	id;
+struct _info {
+    uid_t	id;
     unsigned	flags;
-} specialusers[] = {
-    { /* root */     0,	(FL_KEEPALL|FL_NOAUTOREMOVE)			},
-    { /* ftp  */    11,	(FL_KEEPHOME|FL_NOAUTOADD|FL_NOAUTOREMOVE)	},
-    { /* www-data*/ 33,	(FL_KEEPHOME)					},
-    { /* alias:qmail */   70, (FL_NOAUTOREMOVE)				},
-    { /* qmaild*/   71, (FL_KEEPALL|FL_NOAUTOADD|FL_NOAUTOREMOVE)	}, /* actually only a user */
-    { /* qmails*/   72, (FL_KEEPALL|FL_NOAUTOADD|FL_NOAUTOREMOVE)	}, /* actually only a user */
-    { /* qmailr*/   73, (FL_KEEPALL|FL_NOAUTOADD|FL_NOAUTOREMOVE)	}, /* actually only a user */
-    { /* qmailq*/   74, (FL_KEEPALL|FL_NOAUTOADD|FL_NOAUTOREMOVE)	}, /* actually only a user */
-    { /* qmaill*/   75, (FL_KEEPALL|FL_NOAUTOADD|FL_NOAUTOREMOVE) 	}, /* actually only a user */
-    { /* qmailp*/   76, (FL_KEEPALL|FL_NOAUTOADD|FL_NOAUTOREMOVE)	}, /* actually only a user */
+};
+
+const struct _info specialusers[] = {
+    {  0, (FL_KEEPALL|FL_NOAUTOREMOVE)			},  /* root	*/
+    { 11, (FL_KEEPHOME|FL_NOAUTOADD|FL_NOAUTOREMOVE)	},  /* ftp	*/
+    { 30, (FL_KEEPHOME|FL_NOAUTOADD|FL_NOAUTOREMOVE)	},  /* majordom	*/
+    { 33, (FL_KEEPHOME)					},  /* www-data	*/
+    { 35, (FL_KEEPALL|FL_NOAUTOADD|FL_NOAUTOREMOVE)	},  /* dos	*/
+    { 36, (FL_KEEPALL|FL_NOAUTOADD|FL_NOAUTOREMOVE)	},  /* msql	*/
+    { 41, (FL_KEEPHOME)					},  /* gnats	*/
+    { 70, (FL_NOAUTOREMOVE)				},  /* alias	*/
+    { 71, (FL_KEEPALL|FL_NOAUTOADD|FL_NOAUTOREMOVE)	},  /* qmaild	*/
+    { 72, (FL_KEEPALL|FL_NOAUTOADD|FL_NOAUTOREMOVE)	},  /* qmails	*/
+    { 73, (FL_KEEPALL|FL_NOAUTOADD|FL_NOAUTOREMOVE)	},  /* qmailr	*/
+    { 74, (FL_KEEPALL|FL_NOAUTOADD|FL_NOAUTOREMOVE)	},  /* qmailq	*/
+    { 75, (FL_KEEPALL|FL_NOAUTOADD|FL_NOAUTOREMOVE) 	},  /* qmaill	*/
+    { 76, (FL_KEEPALL|FL_NOAUTOADD|FL_NOAUTOREMOVE)	},  /* qmailp	*/
     { 0, 0}
 };
 
 
+const struct _info specialgroups[] = {
+    {  0, (FL_KEEPALL|FL_NOAUTOREMOVE)			},  /* root	*/
+    { 11, (FL_NOAUTOADD|FL_NOAUTOREMOVE)		},  /* ftp	*/
+    { 31, (FL_NOAUTOADD|FL_NOAUTOREMOVE)		},  /* majordom	*/
+    { 35, (FL_NOAUTOADD|FL_NOAUTOREMOVE)		},  /* dos	*/
+    { 36, (FL_NOAUTOADD|FL_NOAUTOREMOVE)		},  /* msql	*/
+    { 70, (FL_NOAUTOREMOVE)				},  /* qmail	*/
+    { 0, 0}
+};
+
 struct _node {
     union {
 	struct passwd	pw;
@@ -93,9 +105,10 @@ struct _node {
 	t_error
     } t;
     const char*		name;
-    unsigned int	id;
+    uid_t		id;
     struct _node*	next;
-    char		buf[NSS_BUFLEN_PASSWD];
+    struct _node*	last;
+    char		buf[OUR_NSS_BUFSIZE];
 };
 
 const char*	master_passwd	= DEFAULT_PASSWD_MASTER;
@@ -127,6 +140,7 @@ struct _node* create_node() {
     newnode->name=0;
     newnode->id=0;
     newnode->next=NULL;
+    newnode->last=NULL;
     newnode->t=t_error;
 
     return newnode;
@@ -156,7 +170,7 @@ void copy_passwd(struct _node* newnode, const struct _node* node) {
     newnode->d.pw.pw_shell=(newnode->buf+idx);
     idx+=sprintf((newnode->buf+idx), node->d.pw.pw_shell)+1;
 
-    if (idx>NSS_BUFLEN_PASSWD) {
+    if (idx>OUR_NSS_BUFSIZE) {
 	fprintf(stderr, "Aaiieee, we overflowed an entry-buffer, aborting\n");
 	exit(100);
     }
@@ -181,7 +195,7 @@ void copy_shadow(struct _node* newnode, const struct _node* node) {
     newnode->d.sp.sp_pwdp=(newnode->buf+idx);
     idx+=sprintf((newnode->buf+idx), node->d.sp.sp_pwdp)+1;
 
-    if (idx>NSS_BUFLEN_PASSWD) {
+    if (idx>OUR_NSS_BUFSIZE) {
 	fprintf(stderr, "Aaiieee, we overflowed an entry-buffer, aborting\n");
 	exit(100);
     }
@@ -207,7 +221,7 @@ void copy_group(struct _node* newnode, const struct _node* node) {
     /* TODO: properly copy the memberlist. */
     newnode->d.gr.gr_mem=node->d.gr.gr_mem;
 
-    if (idx>NSS_BUFLEN_PASSWD) {
+    if (idx>OUR_NSS_BUFSIZE) {
 	fprintf(stderr, "Aaiieee, we overflowed an entry-buffer, aborting\n");
 	exit(100);
     }
@@ -251,18 +265,12 @@ void add_node(struct _node** head, struct _node* node) {
 
     if (*head==NULL) {
 	*head=node;
+	node->last=node;
 	return;
     }
 
-    if ((*head)->id > node->id) {
-	*head=node;
-	return;
-    }
-
-    for (walk=*head; walk->next && (walk->next->id <= node->id); walk=walk->next)
-	;
-    node->next=walk->next;
-    walk->next=node;
+    (*head)->last->next=node;
+    (*head)->last=node;
 }
 
 
@@ -293,10 +301,10 @@ struct _node* find_by_named_entry(struct _node* head, const struct _node* entry)
 }
 
 
-/* Look for an entry in a list, using the name of _entry as the
+/* Look for an entry in a list, using the id of _entry as the
  * searchkey.
  */
-struct _node* find_by_id(struct _node* head, unsigned int id) {
+struct _node* find_by_id(struct _node* head, uid_t id) {
     while (head) {
 	if (id==head->id)
 	    return head;
@@ -307,23 +315,28 @@ struct _node* find_by_id(struct _node* head, unsigned int id) {
 }
 
 
-/* Function to scan the list of special users to see if a user has a certain
- * flag set.
+/* Function to scan the list of special users or groups to see if a an
+ * entry has a certain flag set.
  */
-int scan_users(unsigned id, unsigned flag) {
-    const struct _userinfo*	walk;
-    for (walk=specialusers; !((walk->id==0) && (walk->flags==0)); walk++)
+int scan_infos(const struct _info *lst, uid_t id, unsigned flag) {
+    const struct _info*	walk;
+    for (walk=lst; !((walk->id==0) && (walk->flags==0)); walk++)
 	if (walk->id==id)
 	    return ((walk->flags&flag)!=0);
     return 0;
 }
 
 /* Just for our convenience */
-int keephome(unsigned id) { return scan_users(id, FL_KEEPHOME); }
-int keepshell(unsigned id) { return scan_users(id, FL_KEEPSHELL); }
-int keepgecos(unsigned id) { return scan_users(id, FL_KEEPGECOS); }
-int noautoremove(unsigned id) { return scan_users(id, FL_NOAUTOREMOVE); }
-int noautoadd(unsigned id) { return scan_users(id, FL_NOAUTOREMOVE); }
+int keephome(const struct _info* lst, uid_t id) {
+    return scan_infos(lst, id, FL_KEEPHOME); }
+int keepshell(const struct _info* lst, uid_t id) {
+    return scan_infos(lst, id, FL_KEEPSHELL); }
+int keepgecos(const struct _info* lst, uid_t id) {
+    return scan_infos(lst, id, FL_KEEPGECOS); }
+int noautoremove(const struct _info* lst, uid_t id) {
+    return scan_infos(lst, id, FL_NOAUTOREMOVE); }
+int noautoadd(const struct _info* lst, uid_t id) {
+    return scan_infos(lst, id, FL_NOAUTOADD); }
 
 /* Function to read passwd database */
 int read_passwd(struct _node** list, const char* file) {
@@ -342,13 +355,13 @@ int read_passwd(struct _node** list, const char* file) {
 
     node=create_node();
 
-    while ((success=fgetpwent_r(input, &(node->d.pw), node->buf, NSS_BUFLEN_PASSWD, &result))==0) {
+    while ((success=fgetpwent_r(input, &(node->d.pw), node->buf, OUR_NSS_BUFSIZE, &result))==0) {
 	node->t=t_passwd;
 	node->name=node->d.pw.pw_name;
 	if (!node->name)
 	    break;
 	if (node->name[0]=='+')
-	    node->id=INT_MAX;
+	    node->id=0;
 	else
 	    node->id=node->d.pw.pw_uid;
 	add_node(list, node);
@@ -384,13 +397,13 @@ int read_group(struct _node** list, const char* file) {
 
     node=create_node();
 
-    while ((success=fgetgrent_r(input, &(node->d.gr), node->buf, NSS_BUFLEN_PASSWD, &result))==0) {
+    while ((success=fgetgrent_r(input, &(node->d.gr), node->buf, OUR_NSS_BUFSIZE, &result))==0) {
 	node->t=t_group;
 	node->name=node->d.gr.gr_name;
 	if (!node->name)
 	    break;
 	if (node->name[0]=='+')
-	    node->id=INT_MAX;
+	    node->id=0;
 	else
 	    node->id=node->d.gr.gr_gid;
 	add_node(list, node);
@@ -398,7 +411,7 @@ int read_group(struct _node** list, const char* file) {
     }
 
     if ((success!=0) && (errno!=ENOENT)) {
-	fprintf(stderr, "Error reading shadow file %s: %s\n", file, strerror(errno));
+	fprintf(stderr, "Error reading group file %s: %s\n", file, strerror(errno));
 	return 2;
     }
     free(node);
@@ -426,7 +439,7 @@ int read_shadow(struct _node** list, const char* file) {
 
     node=create_node();
 
-    while ((success=fgetspent_r(input, &(node->d.sp), node->buf, NSS_BUFLEN_PASSWD, &result))==0) {
+    while ((success=fgetspent_r(input, &(node->d.sp), node->buf, OUR_NSS_BUFSIZE, &result))==0) {
 	node->t=t_shadow;
 	node->id=0;
 	node->name=node->d.sp.sp_namp;
@@ -448,9 +461,57 @@ int read_shadow(struct _node** list, const char* file) {
 }
 
 
-
-/* Simple function to usage information
+/* Small helper functions to safely print strings that might be NULL.
  */
+const char* safestr(const char* str) {
+    if (str==NULL)
+	return "";
+    else
+	return str;
+}
+
+
+/* Implement our own putpwent(3). The version in GNU libc is stupid enough
+ * to not recognize NIS compat entries and will happily turn entry like this:
+ *
+ *    +@staff::::::
+ *
+ * into this:
+ *
+ *    +@staff::0:0:::
+ *
+ */
+
+int fputpwent(const struct passwd *passwd, FILE * f) {
+    int res;
+
+    assert(passwd!=NULL);
+    assert(f!=NULL);
+
+    if (passwd->pw_name[0]=='+')
+	res=fprintf(f, "%s:%s:::%s:%s:%s\n",
+		safestr(passwd->pw_name),
+		safestr(passwd->pw_passwd),
+		safestr(passwd->pw_gecos),
+		safestr(passwd->pw_dir),
+		safestr(passwd->pw_shell));
+    else
+	res=fprintf(f, "%s:%s:%u:%u:%s:%s:%s\n",
+		safestr(passwd->pw_name),
+		safestr(passwd->pw_passwd),
+		passwd->pw_uid, passwd->pw_gid,
+		safestr(passwd->pw_gecos),
+		safestr(passwd->pw_dir),
+		safestr(passwd->pw_shell));
+
+    if (res<0)
+	return -1;
+
+    return 0;
+}
+
+
+/* Simple function to usage information */
 void usage() {
     printf(
 	"Usage: update-passwd [OPTION]...\n"
@@ -490,11 +551,14 @@ void version() {
  * add accounts to shadow here; those will be made automatically at a later
  * stage where we verify the contents of the shadow database
  */
-void process_new_entries(struct _node** passwd, struct _node* master, const char* descr) {
+void process_new_entries(const struct _info* lst, struct _node** passwd, struct _node* master, const char* descr) {
     while (master) {
 	if (find_by_named_entry(*passwd, master)==NULL) {
 	    struct _node* newnode;
 
+	    if (noautoadd(lst, (*passwd)->id))
+		continue;
+
 	    newnode=copy_node(master);
 	    add_node(passwd, newnode);
 	    flag_dirty++;
@@ -511,12 +575,12 @@ void process_new_entries(struct _node** passwd, struct _node* master, const char
  * don't update shadow here since it is verified at a later stage anyway.
  * We will only remove accounts in our range (uids 0-99).
  */
-void process_old_entries(struct _node** passwd, struct _node* master, const char* descr) {
+void process_old_entries(const struct _info* lst, struct _node** passwd, struct _node* master, const char* descr) {
     for (;*passwd; passwd=&((*passwd)->next)) {
 	if (((*passwd)->id<0) || ((*passwd)->id>99))
 	    continue;
 
-	if (noautoremove((*passwd)->id))
+	if (noautoremove(lst, (*passwd)->id))
 	    continue;
 
 	if (find_by_named_entry(master, *passwd)==NULL) {
@@ -563,7 +627,7 @@ void process_changed_accounts(struct _node* passwd, struct _node* master) {
 	    flag_dirty++;
 	}
 
-	if (!keepgecos(passwd->id))
+	if (!keepgecos(specialusers, passwd->id))
 	    if ((passwd->d.pw.pw_gecos==NULL) || (strcmp(passwd->d.pw.pw_gecos, mc->d.pw.pw_gecos)!=0)) {
 		if (opt_verbose)
 		    printf("Changing GECOS of %s to \"%s\".\n", passwd->name, mc->d.pw.pw_gecos);
@@ -575,10 +639,10 @@ void process_changed_accounts(struct _node* passwd, struct _node* master) {
 		flag_dirty++;
 	    }
 
-	if (!keephome(passwd->id))
+	if (!keephome(specialusers, passwd->id))
 	    if ((passwd->d.pw.pw_dir==NULL) || (strcmp(passwd->d.pw.pw_dir, mc->d.pw.pw_dir)!=0)) {
 		if (opt_verbose)
-		    printf("Changing homedirectory of %s to %s\n", passwd->name, mc->d.pw.pw_dir);
+		    printf("Changing home-directory of %s to %s\n", passwd->name, mc->d.pw.pw_dir);
 		/* We update the pw_dir entry of passwd so it now points into the
 		 * buffer from mc. This is safe for us, since we know we won't free
 		 * the data in mc until after we are done.
@@ -587,7 +651,7 @@ void process_changed_accounts(struct _node* passwd, struct _node* master) {
 		flag_dirty++;
 	    }
 
-	if (!keepshell(passwd->id))
+	if (!keepshell(specialusers, passwd->id))
 	    if ((passwd->d.pw.pw_shell==NULL) || (strcmp(passwd->d.pw.pw_shell, mc->d.pw.pw_shell)!=0)) {
 		if (opt_verbose)
 		    printf("Changing shell of %s to %s\n", passwd->name, mc->d.pw.pw_shell);
@@ -640,7 +704,7 @@ int write_passwd(const struct _node* passwd, const char* file) {
 
     for (;passwd; passwd=passwd->next) {
 	assert(passwd->t==t_passwd);
-	if (putpwent(&(passwd->d.pw), output)!=0) {
+	if (fputpwent(&(passwd->d.pw), output)!=0) {
 	    fprintf(stderr, "Error writing passwd-entry: %s\n", strerror(errno));
 	    return 0;
 	}
@@ -767,7 +831,7 @@ int copy_filemodes(const char* source, const char* target) {
 	return 0;
     }
 
-    if (lchown(target, st.st_uid, st.st_gid)!=0)
+    if (lchown(target, st.st_uid, st.st_gid)!=0) {
 	/* Hmm, this failed. Lets try a normal chown in case we
 	 * are running on a kernel that doesn't support lchown
 	 */
@@ -791,6 +855,7 @@ int copy_filemodes(const char* source, const char* target) {
 		    source, strerror(errno));
 	    return 0;
 	}
+    }
 
     return 1;
 }
@@ -832,24 +897,35 @@ int replace_file(const char* org, const char* new, const char* backup) {
  * new copy, since it's useless anyway.
  */
 int put_file_in_place(const char* source, const char* target) {
-    char	uf[PATH_MAX];
+    char*	uf;
+    int		ret;
 
     if (opt_verbose>1)
 	printf("Replacing \"%s\" with \"%s\"\n", target, source);
 
-    snprintf(uf, PATH_MAX, "%s%s", target, BACKUP_EXTENSION);
+    asprintf(&uf, "%s%s", target, BACKUP_EXTENSION);
 
-    if (!copy_filemodes(target, source))
+    if (uf==NULL) {
+	fprintf(stderr, "Not enough memory available\n");
 	return 0;
+    }
 
-    return replace_file(target, source, uf);
+    if (!copy_filemodes(target, source)) {
+	free(uf);
+	return 0;
+    }
+
+    ret=replace_file(target, source, uf);
+    free(uf);
+
+    return ret;
 }
 
 
 /* Rewrite the account-database if we made any changes
  */
 int commit_files() {
-    char	wf[PATH_MAX];
+    char*	wf;
 
     if (!flag_dirty) {
 	if (opt_verbose)
@@ -864,25 +940,64 @@ int commit_files() {
 
     printf("%d changes have been made, rewriting files\n", flag_dirty);
 
-    snprintf(wf, PATH_MAX, "%s%s", sys_passwd, WRITE_EXTENSION);
-    if (!write_passwd(system_accounts, wf))
-	return 0;
-    if (!put_file_in_place(wf, sys_passwd))
-	return 0;
+    asprintf(&wf, "%s%s", sys_passwd, WRITE_EXTENSION);
 
-    if (system_shadow!=NULL) {
-	snprintf(wf, PATH_MAX, "%s%s", sys_shadow, WRITE_EXTENSION);
-	if (!write_shadow(system_shadow, wf))
-	    return 0;
-	if (!put_file_in_place(wf, sys_shadow))
-	    return 0;
+    if (wf==NULL) {
+	fprintf(stderr, "Not enough memory available\n");
+	return 0;
     }
 
-    snprintf(wf, PATH_MAX, "%s%s", sys_group, WRITE_EXTENSION);
-    if (!write_group(system_groups, wf))
+    if (!write_passwd(system_accounts, wf)) {
+	free(wf);
 	return 0;
-    if (!put_file_in_place(wf, sys_group))
+    }
+
+    if (!put_file_in_place(wf, sys_passwd)) {
+	free(wf);
 	return 0;
+    }
+
+    free(wf);
+
+    if (system_shadow!=NULL) {
+	asprintf(&wf, "%s%s", sys_shadow, WRITE_EXTENSION);
+
+	if (wf==NULL) {
+	    fprintf(stderr, "Not enough memory available\n");
+	    return 0;
+	}
+
+	if (!write_shadow(system_shadow, wf)) {
+	    free(wf);
+	    return 0;
+	}
+
+	if (!put_file_in_place(wf, sys_shadow)) {
+	    free(wf);
+	    return 0;
+	}
+
+	free(wf);
+    }
+
+    asprintf(&wf, "%s%s", sys_group, WRITE_EXTENSION);
+
+    if (wf==NULL) {
+	fprintf(stderr, "Not enough memory available\n");
+	return 0;
+    }
+
+    if (!write_group(system_groups, wf)) {
+	free(wf);
+	return 0;
+    }
+
+    if (!put_file_in_place(wf, sys_group)) {
+	free(wf);
+	return 0;
+    }
+
+    free(wf);
 
     return 1;
 }
@@ -989,12 +1104,12 @@ int main(int argc, char** argv) {
     if (read_group(&system_groups, sys_group)!=0)
 	return 2;
 
-    process_new_entries(&system_accounts, master_accounts, "user");
-    process_old_entries(&system_accounts, master_accounts, "user");
+    process_new_entries(specialusers, &system_accounts, master_accounts, "user");
+    process_old_entries(specialusers, &system_accounts, master_accounts, "user");
     process_changed_accounts(system_accounts, master_accounts);
 
-    process_new_entries(&system_groups, master_groups, "group");
-    process_old_entries(&system_groups, master_groups, "group");
+    process_new_entries(specialgroups, &system_groups, master_groups, "group");
+    process_old_entries(specialgroups, &system_groups, master_groups, "group");
     process_changed_groups(system_groups, master_groups);
 
     if (opt_sanity)
